@@ -20,10 +20,11 @@
 Supported drivers, in order of preference:
 - pyodbc (recommended, others are not well tested)
 - adodbapi
-
 """
+
 import os
 import sys
+import shutil
 
 from logilab import database as db
 from logilab.database.sqlserver import _PyodbcAdapter, _AdodbapiAdapter
@@ -79,16 +80,6 @@ class _SqlServer2005FuncHelper(db._GenericAdvFuncHelper):
         cursor.execute(sql)
         return [r[0] for r in cursor.fetchall()]
 
-    def list_indices(self, cursor, table=None):
-        """return the list of indices of a database, only for the given table if specified"""
-        sql = "SELECT name FROM sys.indexes"
-        if table:
-            sql = ("SELECT ind.name FROM sys.indexes as ind, sys.objects as obj WHERE "
-                   "obj.object_id = ind.object_id AND obj.name = '%s'"
-                   % table)
-        cursor.execute(sql)
-        return [r[0] for r in cursor.fetchall()]
-
     def binary_value(self, value):
         return StringIO.StringIO(value)
 
@@ -106,6 +97,34 @@ class _SqlServer2005FuncHelper(db._GenericAdvFuncHelper):
                  dbname or self.dbname, backupfile],
                 ]
 
+    def _index_names(self, cursor, table, column):
+        """
+        return the list of index_information for table.column
+        index_information is a tuple:
+        (name, index_type, is_unique, is_unique_constraint)
+
+        See http://msdn.microsoft.com/en-us/library/ms173760.aspx for more
+        information
+        """
+        has_index_sql = """\
+SELECT i.name AS index_name,
+       i.type_desc,
+       i.is_unique,
+       i.is_unique_constraint
+FROM sys.indexes AS i, sys.index_columns as j, sys.columns as k
+WHERE is_hypothetical = 0 AND i.index_id <> 0
+AND i.object_id = j.object_id
+AND i.index_id = j.index_id
+AND i.object_id = OBJECT_ID('%(table)s')
+AND k.name = '%(col)s'
+AND k.object_id=i.object_id
+AND j.column_id = k.column_id;"""
+        cursor.execute(has_index_sql % {'table': table, 'col': column})
+        return cursor.fetchall()
+
+    def index_exists(self, cursor, table, column, unique=False):
+        indexes = self._index_names(cursor, table, column)
+        return len(indexes) > 0
 
     def sql_concat_string(self, lhs, rhs):
         return '%s + %s' % (lhs, rhs)
@@ -116,22 +135,45 @@ class _SqlServer2005FuncHelper(db._GenericAdvFuncHelper):
         return "CREATE TABLE %s (%s);" % (table_name, table_schema)
 
     def sql_change_col_type(self, table, column, coltype, null_allowed):
-        if null_allowed:
-            cmd = 'ALTER TABLE %s ALTER COLUMN %s %s NULL'
-        else:
-            cmd = 'ALTER TABLE %s ALTER COLUMN %s %s NOT NULL'
-        return cmd % (table, column, coltype)
+        raise NotImplementedError('use .change_col_type()')
 
     def sql_set_null_allowed(self, table, column, coltype, null_allowed):
-        return self.sql_change_col_type(table, column, coltype, null_allowed)
+        raise NotImplementedError('use .set_null_allowed()')
+
+    def change_col_type(self, cursor, table, column, coltype, null_allowed):
+        alter = []
+        drops = []
+        creates = []
+        print "change col type for %s.%s to %s %s" % (table, column, coltype, null_allowed and 'NULL' or 'NOT NULL')
+
+        for idx_name, idx_type, is_unique, is_unique_cstr in self._index_names(cursor, table, column):
+            if is_unique_cstr:
+                drops.append('ALTER TABLE %s DROP CONSTRAINT %s' % (table, idx_name))
+                creates.append('ALTER TABLE %s ADD CONSTRAINT %s UNIQUE (%s)' % (table, idx_name, column))
+            else:
+                drops.append('DROP INDEX %s ON %s' % (idx_name, table))
+                if is_unique:
+                    unique = 'UNIQUE'
+                else:
+                    unique = ''
+                creates.append('CREATE %s %s INDEX %s ON %s(%s)' % (unique, idx_type, idx_name, table, column))
+
+        if null_allowed:
+            null = 'NULL'
+        else:
+            null = 'NOT NULL'
+        alter.append('ALTER TABLE %s ALTER COLUMN %s %s %s' % (table, column, coltype, null))
+        for stmt in drops + alter + creates:
+            cursor.execute(stmt)
+
+    def set_null_allowed(self, cursor, table, column, coltype, null_allowed):
+        return self.change_col_type(cursor, table, column, coltype, null_allowed)
 
     def temporary_table_name(self, table_name):
         if not table_name.startswith('#'):
             table_name = '#' + table_name
         return table_name
 
-
-
     @staticmethod
     def _do_backup():
         import time
@@ -139,16 +181,23 @@ class _SqlServer2005FuncHelper(db._GenericAdvFuncHelper):
         dbhost = sys.argv[2]
         dbname = sys.argv[3]
         filename = sys.argv[4]
-        cnx = get_connection(driver='sqlserver2005', host=dbhost, database=dbname, extra_args='autocommit;trusted_connection')
+        cnx = get_connection(driver='sqlserver2005',
+                             host=dbhost, database=dbname,
+                             extra_args='autocommit;trusted_connection')
         cursor = cnx.cursor()
-        cursor.execute("BACKUP DATABASE %(db)s TO DISK= %(path)s ", {'db':dbname, 'path':filename,})
+        sql_server_local_filename = r"C:\Backups\%s" % dbname
+        file_share_filename = r"\\%s\Backups\%s" % (dbhost, dbname)
+        cursor.execute("BACKUP DATABASE %(db)s TO DISK= %(path)s ",
+                       {'db':dbname,
+                        'path':sql_server_local_filename,
+                        })
         prev_size = -1
         err_count = 0
         same_size_count = 0
         while err_count < 10 and same_size_count < 10:
             time.sleep(1)
             try:
-                size = os.path.getsize(filename)
+                size = os.path.getsize(file_share_filename)
             except OSError, exc:
                 err_count += 1
                 print exc
@@ -157,6 +206,8 @@ class _SqlServer2005FuncHelper(db._GenericAdvFuncHelper):
                 prev_size = size
             else:
                 same_size_count += 1
+        shutil.copy(file_share_filename, filename)
+        os.remove(file_share_filename)
         cnx.close()
         sys.exit(0)
 
@@ -167,18 +218,30 @@ class _SqlServer2005FuncHelper(db._GenericAdvFuncHelper):
         dbhost = sys.argv[2]
         dbname = sys.argv[3]
         filename = sys.argv[4]
-        cnx = get_connection(driver='sqlserver2005', host=dbhost, database='master', extra_args='autocommit;trusted_connection')
+        sql_server_local_filename = r"C:\Backups\%s" % dbname
+        file_share_filename = r"\\%s\Backups\%s" % (dbhost, dbname)
+        shutil.copy(filename, file_share_filename)
+        cnx = get_connection(driver='sqlserver2005',
+                             host=dbhost, database='master',
+                             extra_args='autocommit;trusted_connection')
+
         cursor = cnx.cursor()
-        cursor.execute("RESTORE DATABASE %(db)s FROM DISK= %(path)s WITH REPLACE", {'db':dbname, 'path':filename,})
+        cursor.execute("RESTORE DATABASE %(db)s FROM DISK= %(path)s WITH REPLACE",
+                       {'db':dbname,
+                        'path':sql_server_local_filename,
+                        })
         import time
         sleeptime = 10
         while True:
             time.sleep(sleeptime)
             try:
-                cnx = get_connection(driver='sqlserver2005', host=dbhost, database=dbname, extra_args='trusted_connection')
+                cnx = get_connection(driver='sqlserver2005',
+                                     host=dbhost, database=dbname,
+                                     extra_args='trusted_connection')
                 break
             except:
                 sleeptime = min(sleeptime*2, 300)
+        os.remove(file_share_filename)
         sys.exit(0)
 
 
